;;;; -*- lisp -*-

(in-package :it.bese.tar)

(defmethod record.data ((rec tar-file-record))
  (with-slots (data extract-from offset size) rec
    (when (and (null data) extract-from)
      ;; no data in the record and we have a file to extract the data
      ;; from, read it.
      (with-open-file (rec-stream extract-from :direction :input :element-type '(unsigned-byte 8))
        (file-position rec-stream offset)
        (multiple-value-bind (buf pos)
            (read-sequence* (make-array size :element-type '(unsigned-byte 8))
                            rec-stream)
	  (assert (= pos size)
		  (rec-stream)
		  "Unable to read ~S bytes from ~S (offset: ~S)" size rec-stream offset)
	  (setf (slot-value rec 'data) buf))))
    data))

(defun read-tar-file (filename &optional (extract-p t))
  (with-open-file (tar filename
		       :direction :input
		       :element-type '(unsigned-byte 8))
    (assert (not (zerop (file-length tar)))
	    (tar)
	    "Um, so ~S is a tar, but it's empty...yeah, right." filename)
    (let ((files (loop
                    for record = (read-tar-record tar extract-p)
                    while record
                    do (setf (record.extract-file record) (truename filename)
                             (record.dirty-p record) nil)
                    collect record)))
      ;; ok, read all the files, now make sure there's one more empty
      ;; zero block (header-tar-header read the second ot last one
      (multiple-value-bind (buf pos)
          (read-sequence* (make-array 512 :element-type '(unsigned-byte 8)) tar)
	(if (and (= pos 512)
		 (every #'zerop buf))
	    (return-from read-tar-file files)
	    (progn
	      (warn "Tar file ended before last empty block. Ignoring #D bytes of junk." pos)
	      (return-from read-tar-file files)))))))

(defun read-tar-record (stream extract-p)
  (let ((file (read-tar-record-header stream)))
    (when (null file)
      (return-from read-tar-record nil))
    (let ((buffer-size (* 512 (ceiling (record.size file) 512))))
      (if extract-p
	  ;; need to also extract the file contents
          (multiple-value-bind (buf pos)
              (read-sequence* (make-array buffer-size :element-type '(unsigned-byte 8)) stream)
	    (assert (= pos buffer-size)
		    (stream)
		    "Stream ended before reading record header.")
	    (setf (record.data file) (make-array (record.size file)
                                                 :element-type '(unsigned-byte 8)
                                                 :displaced-to buf)))
	  ;; don't need to read the file contents, move the file-position forward
	  (assert (file-position stream (+ (file-position stream) buffer-size))
		  (stream)
		  "For some odd reason we can't move the file-position forward")))
    file))

(defun read-tar-record-header (stream)
  "Read a TAR header from STREAM."
  (multiple-value-bind (header pos)
      (read-sequence* (make-array 512 :element-type '(unsigned-byte 8)) stream)
    (assert (= 512 pos)
	    (pos)
	    "File ended before reading complete header.")
    (when (every #'zerop header)
      ;; last block, no header here
      (return-from read-tar-record-header nil))
    (make-tar-record-from-header (file-position stream) header)))

(defmacro parse-octalf (octal-string)
  `(setf ,octal-string (parse-octal ,octal-string)))

(defun make-tar-record-from-header (stream-offset header-buffer)
  (let ((record (make-instance 'tar-file-record
                               :offset stream-offset
                               :name     (string-from-array header-buffer :start 0   :end 99)
                               :mode     (string-from-array header-buffer :start 100 :end 107)
                               :uid      (string-from-array header-buffer :start 108 :end 115)
                               :gid      (string-from-array header-buffer :start 116 :end 123)
                               :size     (string-from-array header-buffer :start 124 :end 135)
                               :mtime    (string-from-array header-buffer :start 136 :end 147)
                               :chksum   (string-from-array header-buffer :start 148 :end 155)
                               :typeflag (string-from-array header-buffer :start 156 :end 156)
                               :linkname (string-from-array header-buffer :start 157 :end 256)
                               :magic    (string-from-array header-buffer :start 257 :end 262)
                               :version  (string-from-array header-buffer :start 263 :end 264)
                               :uname    (string-from-array header-buffer :start 265 :end 296)
                               :gname    (string-from-array header-buffer :start 297 :end 328)
                               :devmajor (string-from-array header-buffer :start 329 :end 336)
                               :devminor (string-from-array header-buffer :start 337 :end 344)
                               :prefix   (string-from-array header-buffer :start 345 :end 499))))
    (with-slots (mode uid gid size mtime chksum magic)
        record
      (parse-octalf mode)
      (parse-octalf uid)
      (parse-octalf gid)
      (setf size (or (parse-octal size)
                     0))
      (parse-octalf mtime)
      (parse-octalf chksum)
      (if (and (< 0 (length magic))
	       (string= "ustar" (subseq magic 0 5)))
	  (post-process-ustar record)
          (post-process-standard record)))))

(defun parse-octal (octal-string)
  ;; trim space chars from the start and end of octal string
  (let ((trimmed (trim-string octal-string #\Space)))
    (if (every #'digit-char-p trimmed)
	(parse-integer trimmed :radix 8)
        nil)))

(defun post-process-standard (record)
  (with-slots (typeflag) record
    (setf typeflag (switch (typeflag :test #'string=)
                     ("l" :link)
                     ("s" :symbolic)
                     ("o" :regular)
		     (t :regular)))
    record))

(defun post-process-ustar (record)
  (with-slots (name prefix typeflag version devminor devmajor) record
    (setf name (concatenate 'string prefix name))
    (setf typeflag (eswitch (typeflag :test #'string=)
                     (("0" "") :regular)
                     ("1" :link)
                     ("2" :symbolic)
                     ("3" :char-special)
                     ("4" :block-special)
                     ("5" :directory)
                     ("6" :fifo-special)))
    (parse-octalf version)
    (parse-octalf devmajor)
    (parse-octalf devminor)
    record))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; Copyright (c) 2002-2003, Edward Marco Baringer
;;;; All rights reserved. 
;;;; 
;;;; Redistribution and use in source and binary forms, with or without
;;;; modification, are permitted provided that the following conditions are
;;;; met:
;;;; 
;;;;  - Redistributions of source code must retain the above copyright
;;;;    notice, this list of conditions and the following disclaimer.
;;;; 
;;;;  - Redistributions in binary form must reproduce the above copyright
;;;;    notice, this list of conditions and the following disclaimer in the
;;;;    documentation and/or other materials provided with the distribution.
;;;;
;;;;  - Neither the name of Edward Marco Baringer, nor BESE, nor the names
;;;;    of its contributors may be used to endorse or promote products
;;;;    derived from this software without specific prior written permission.
;;;; 
;;;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
;;;; "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;;;; LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
;;;; A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
;;;; OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
;;;; SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
;;;; LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
;;;; DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
;;;; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
;;;; (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
;;;; OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
